@@ -159,6 +159,11 @@ NO_STDIN = os.isatty(0)
 NO_STDOUT = sys.stdout.isatty()
 
 
+# Instantiate logger facilities for unit testing
+log = logging.getLogger('testing')
+log.setLevel(logging.CRITICAL)
+
+
 def _configure_logging(parsed_args: argparse.Namespace) -> ():
     """
     Prepare and return two logging objects: one for the general program, and
@@ -200,7 +205,7 @@ def _configure_logging(parsed_args: argparse.Namespace) -> ():
     return log, meraki_log
 
 
-def _cmd_title(mtdstr: str) -> '':
+def _cmd_title(mtdstr: str) -> str:
     """
     Generate command/method title, based on its name, which is
         displayed in the list of commands under a type (class).
@@ -213,7 +218,7 @@ def _cmd_title(mtdstr: str) -> '':
     return ' '.join(wordlist).title()  # Camel case all words and join them
 
 
-def _cmd_help(arg_obj: Args) -> '':
+def _cmd_help(arg_obj: Args) -> str:
     """
     Generate the description/instructions section in a command/method's help
         text. Use the docstring of the method and add the method's signature
@@ -280,8 +285,9 @@ def _get_method_params(parsed_args: argparse.Namespace,
                 log.debug(f'Loaded kwargs:\n{json.dumps(kwargs, indent=4)}')
                 arg_dict.update(kwargs)  # Add the data to the keyword values
             except json.decoder.JSONDecodeError as e:
-                log.critical(f'Error loading JSON kwargs. Check syntax.')
+                log.critical('Error loading JSON kwargs. Check syntax.')
                 log.exception(e)
+                sys.exit()
     return (positionals, arg_dict)
 
 
@@ -314,25 +320,23 @@ def _translate_input(arg_dicts: list, translation_strs: list) -> []:
     return result
 
 
-def _get_stdin_args(parsed_args: argparse.Namespace, arg_obj: Args) -> []:
+def _json_to_args(json_string: str, parsed_args: argparse.Namespace,
+                  arg_obj: Args) -> []:
     """
-    Get input objects from a pipeline (STDIN), run them through a translator
-        if the user provided key translations, reconcile them with CLI inputs
-        from the user (preferring arguments from the user) and conform the
-        arguments to the parameters required by the target method.
+    Parse a raw JSON string into method-ready arguments. Parse JSON into
+        objects, run them through a translator if the user provided key
+        translations, reconcile them with CLI inputs from the user (preferring
+        arguments from the user) and conform the arguments to the parameters
+        required by the target method.
 
+    - json_string: String of valid JSON objects. Example:
+        '[{"vlanId": "100"}]'
     - parsed_args: Parsed arguments namespace object from argparse.
     - arg_obj: Instance of the Args() class which holds parameter info for the
         target method.
     """
     result = []
-    stdin = sys.stdin.read()  # Get the raw STDIN data
-    log.debug(f'Raw STDIN data:\n{stdin}')
-    if not stdin:  # If the STDIN data is empty
-        log.critical(f'STDIN is empty. Source instance returned no data. '
-                     'Exiting program.')
-        sys.exit()
-    stdin_arg_dicts = json.loads(stdin)  # Interpret as JSON
+    stdin_arg_dicts = json.loads(json_string)  # Interpret as JSON
     log.debug(f'Loaded STDIN data: {stdin_arg_dicts}')
     if type(stdin_arg_dicts) is not list:  # Make a list if not a list
         stdin_arg_dicts = [stdin_arg_dicts]
@@ -363,6 +367,26 @@ def _get_stdin_args(parsed_args: argparse.Namespace, arg_obj: Args) -> []:
     return result
 
 
+def _get_stdin_args(parsed_args: argparse.Namespace, arg_obj: Args) -> []:
+    """
+    Get input objects from a pipeline (STDIN), run them through a translator
+        if the user provided key translations, reconcile them with CLI inputs
+        from the user (preferring arguments from the user) and conform the
+        arguments to the parameters required by the target method.
+
+    - parsed_args: Parsed arguments namespace object from argparse.
+    - arg_obj: Instance of the Args() class which holds parameter info for the
+        target method.
+    """
+    stdin = sys.stdin.read()  # Get the raw STDIN data
+    log.debug(f'Raw STDIN data:\n{stdin}')
+    if not stdin:  # If the STDIN data is empty
+        log.critical('STDIN is empty. Source instance returned no data. '
+                     'Exiting program.')
+        sys.exit()
+    return _json_to_args(stdin, parsed_args, arg_obj)
+
+
 def _print_commands(parsed_args: argparse.Namespace,
                     argtups: tuple, arg_obj: Args) -> None:
     """
@@ -522,7 +546,7 @@ def _nice_table(listofdicts: list) -> None:
         [{'id': '1', 'name': 'THING1'}, {'id': '2', 'name': 'THING2'}]
     """
     if not listofdicts:
-        log.warning(f'Empty table data. Exiting.')
+        log.warning('Empty table data. Exiting.')
         sys.exit()
     console = Console()  # Instantiate the Rich console
     # Instantiate the table object with header settings
@@ -553,32 +577,25 @@ def _output(parsed_args: argparse.Namespace, result: list) -> None:
     #     negative. It is confusing but unavoidable.
     if parsed_args.jsonOutput or not NO_STDOUT:
         log.debug(f'Outputting to JSON. '
-                  f'Forward pipelining: {NO_STDOUT}')
+                  f'Forward pipelining: {not NO_STDOUT}')
         print(json.dumps(result, indent=4))
     else:  # Otherwise let's try to output a nice table.
-        try:  # Catch any exceptions in tabulating the data
-            # Make data a list if it is not already
-            if type(result) is dict:
-                result = [result]
-            # Prepare data to be formatted into a table by removing value
-            #     types unfriendly to tabulation.
-            tabledicts = _table_ready_dicts(result)
-            # If the user has defined column filtering/ordering
-            if parsed_args.columns:
-                tabledicts = _column_filter(tabledicts, parsed_args.columns)
-            _nice_table(tabledicts)  # Output our nice table to the CLI
-        except Exception as e:  # If we could not tabulate the results
-            # Alert the user
-            log.info('Exception thrown in table formatting. Using JSON. '
-                     'Logging exception at debug level.')
-            if log.getEffectiveLevel() <= logging.DEBUG:
-                log.exception(e)
-            print(json.dumps(result, indent=4))
+        # Make data a list if it is not already
+        if type(result) is dict:
+            result = [result]
+        # Prepare data to be formatted into a table by removing value
+        #     types unfriendly to tabulation.
+        tabledicts = _table_ready_dicts(result)
+        # If the user has defined column filtering/ordering
+        if parsed_args.columns:
+            tabledicts = _column_filter(tabledicts, parsed_args.columns)
+        _nice_table(tabledicts)  # Output our nice table to the CLI
 
 
-def main() -> None:
+def main(argstring=None) -> None:
     """
-    Primary function called from native script.
+    Primary function called from native script. Allow an argument string to
+        be passed in for testing purposes.
     """
     # Instantiate a fake instance of the API so we can parse it and build the
     #     argparse structure from it
@@ -595,7 +612,7 @@ def main() -> None:
                              help='Show help at any command level',
                              action='help')
     basic_group.add_argument('-v', '--version',
-                             version=f'Meraki-CLI v{__version__.version} | '
+                             version=f'Meraki-CLI v{__version__} | '
                              f'Meraki API Library v{meraki.__version__}',
                              action="version")
     basic_group.add_argument('-k', '--apiKey',
@@ -684,17 +701,17 @@ def main() -> None:
             parser_map[clsstr][mtdstr] = arg_obj
             # Iterate the positionals and create argument options for each
             for arg in arg_obj.positionals:
-                param = cmd.add_argument(f'--{arg.name}',
-                                         dest=arg.name,
-                                         help='(required)',
-                                         # Don't require the arg if we are
-                                         #     parsing the STDIN data
-                                         required=NO_STDIN,
-                                         # Use the annotation map to pass
-                                         #     additional params to the
-                                         #     argument option to make it
-                                         #     easier to use and read
-                                         **ANNOTATION_MAP[arg.annotation])
+                cmd.add_argument(f'--{arg.name}',
+                                 dest=arg.name,
+                                 help='(required)',
+                                 # Don't require the arg if we are
+                                 #     parsing the STDIN data
+                                 required=NO_STDIN,
+                                 # Use the annotation map to pass
+                                 #     additional params to the
+                                 #     argument option to make it
+                                 #     easier to use and read
+                                 **ANNOTATION_MAP[arg.annotation])
                 # Add the parameter name to the positionals list in the map
             for arg in arg_obj.keywords:
                 cmd.add_argument(f'--{arg.name}',
@@ -706,9 +723,14 @@ def main() -> None:
             if arg_obj.varkw:
                 cmd.add_argument(f'--{arg_obj.varkw.name}',
                                  dest=arg_obj.varkw.name,
-                                 help=f'(JSON formatted extra arguments)',
+                                 help='(JSON formatted extra arguments)',
                                  metavar='JSON_STRING',)
-    args = parser.parse_args()  # Parse the user provided CLI commands/args
+    # If an argstring was passed in, we are probably being tested
+    if argstring:
+        # Split up the args and pass them in as a list to be parsed
+        args = parser.parse_args(argstring.split(' '))
+    else:
+        args = parser.parse_args()  # Parse the user provided CLI commands/args
     global log  # Make the "log" variable global to anybody can use it
     # Pull in the two logging systems
     log, meraki_log = _configure_logging(args)
@@ -739,7 +761,7 @@ def main() -> None:
     target_method = getattr(getattr(api, args.type), args.command)
     log.info(f'Target method is {target_method}')
     if NO_STDIN:  # If not in front of a pipe (not being fed by STDIN)
-        log.info(f'No STDIN detected. No pipe behind this instance.')
+        log.info('No STDIN detected. No pipe behind this instance.')
         # Remove the static args and get a clean dict back only containing
         #     argument meant to become parameters in the target method call
         arg_dict = _clean_args(args)
@@ -753,7 +775,7 @@ def main() -> None:
         argtups = [_get_method_params(args, arg_dict,
                                       parser_map[args.type][args.command])]
     else:  # If we are being fed by STDIN through a pipeline
-        log.info(f'STDIN detected. Pipe exists behind this instance.')
+        log.info('STDIN detected. Pipe exists behind this instance.')
         # Pull a list of tuples, each containing the parameters tailored for a
         #     call against the target method.
         # Tuple should have a list of positional arguments in [0] and a dict
@@ -764,7 +786,7 @@ def main() -> None:
     # If the user threw the switch to output commands instead of execute
     #     them
     if args.output_commands:
-        log.debug(f'Outputting piped args as commands and cancelling '
+        log.debug('Outputting piped args as commands and cancelling '
                   'execution.')
         # Print the commands to STDOUT and exit the program
         _print_commands(args, argtups, parser_map[args.type][args.command])
